from abc import abstractmethod
from typing import Any

from django_microservice_common.utils.singletone import Singleton


class BaseDependencyContainer:
    def __init__(self):
        self.dependencies = {}

    @abstractmethod
    def register_dependency(self, dependency_name: str, dependency: Any):
        """
        Abstraction of register a new dependency into container
        Args:
            dependency_name (str): A unique name for the dependency.
            dependency (object): The dependency object to be registered.
        """

    @abstractmethod
    def resolve_dependency(self, dependency_name: str):
        """
        Abstraction of accessing an existed dependency from container
        Args:
            dependency_name (str): The name of the dependency to resolve.
        Returns:
            object: The resolved dependency object.
        """


class ApplicationDependencyContainer(Singleton, BaseDependencyContainer):

    def register_dependency(self, dependency_name: str, dependency: Any):
        self.dependencies[dependency_name.lower()] = dependency

    def resolve_dependency(self, dependency_name: str):
        if dependency_name in self.dependencies:
            return self.dependencies[dependency_name.lower()]
        else:
            raise Exception(f"Dependency '{dependency_name}' not found in the container.")



from typing import TypeVar

from django.db import models

# Generic type for a Django model
# Reference: https://mypy.readthedocs.io/en/stable/kinds_of_types.html#the-type-of-class-objects
DjangoModelType = TypeVar('DjangoModelType', bound=models.Model)



from typing import List, Dict, Any, Tuple

from ..types.types import DjangoModelType


def model_update(
        *,
        instance: DjangoModelType,
        fields: List[str],
        data: Dict[str, Any]
) -> Tuple[DjangoModelType, bool]:
    """
    Generic update service meant to be reused in local update services

    For example:

    def user_update(*, user: User, data) -> User:
        fields = ['first_name', 'last_name']
        user, has_updated = model_update(instance=user, fields=fields, data=data)

        // Do other actions with the user here

        return user

    Return value: Tuple with the following elements:
        1. The instance we updated
        2. A boolean value representing whether we performed an update or not.
    """
    has_updated = False

    for field in fields:
        # Skip if a field is not present in the actual data
        if field not in data:
            continue

        if getattr(instance, field) != data[field]:
            has_updated = True
            setattr(instance, field, data[field])

    # Perform an update only if any of the fields was actually changed
    if has_updated:
        instance.full_clean()
        # Update only the fields that are meant to be updated.
        # Django's docs reference:
        # https://docs.djangoproject.com/en/dev/ref/models/instances/#specifying-which-fields-to-save
        instance.save(update_fields=fields)

    return instance, has_updated



from django.conf import settings
from django.shortcuts import get_object_or_404
from django.http import Http404
from django.core.exceptions import ImproperlyConfigured

from rest_framework import serializers


def make_mock_object(**kwargs):
    return type("", (object,), kwargs)


def get_object(model_or_queryset, **kwargs):
    """
    Reuse get_object_or_404 since the implementation supports both Model && queryset.
    Catch Http404 & return None
    """
    try:
        return get_object_or_404(model_or_queryset, **kwargs)
    except Http404:
        return None


def assert_settings(required_settings, error_message_prefix=""):
    """
    Checks if each item from `required_settings` is present in Django settings
    """
    not_present = []
    values = {}

    for required_setting in required_settings:
        if not hasattr(settings, required_setting):
            not_present.append(required_setting)
            continue

        values[required_setting] = getattr(settings, required_setting)

    if not_present:
        if not error_message_prefix:
            error_message_prefix = "Required settings not found."

        stringified_not_present = ", ".join(not_present)

        raise ImproperlyConfigured(f"{error_message_prefix} Could not find: {stringified_not_present}")

    return values



class Singleton:
    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super(Singleton, cls).__new__(cls)
        return cls.__instance



from collections import OrderedDict

from rest_framework.pagination import LimitOffsetPagination as _LimitOffsetPagination
from rest_framework.response import Response


def get_paginated_response(*, pagination_class, serializer_class, queryset, request, view):
    paginator = pagination_class()

    page = paginator.paginate_queryset(queryset, request, view=view)

    if page is not None:
        serializer = serializer_class(page, many=True)
        return paginator.get_paginated_response(serializer.data)

    serializer = serializer_class(queryset, many=True)

    return Response(data=serializer.data)

def get_paginated_response_context(*, pagination_class, serializer_class, queryset, request, view):
    paginator = pagination_class()

    page = paginator.paginate_queryset(queryset, request, view=view)

    if page is not None:
        serializer = serializer_class(page, many=True, context={'request':request})
        return paginator.get_paginated_response(serializer.data)

    serializer = serializer_class(queryset, many=True, context={'request':request})

    return Response(data=serializer.data)


class LimitOffsetPagination(_LimitOffsetPagination):
    default_limit = 10
    max_limit = 50

    def get_paginated_data(self, data):
        return OrderedDict([
            ('limit', self.limit),
            ('offset', self.offset),
            ('count', self.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ])

    def get_paginated_response(self, data):
        """
        We redefine this method in order to return `limit` and `offset`.
        This is used by the frontend to construct the pagination itself.
        """
        return Response(OrderedDict([
            ('limit', self.limit),
            ('offset', self.offset),
            ('count', self.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))



from djangorestframework_camel_case.parser import CamelCaseFormParser, CamelCaseMultiPartParser, CamelCaseJSONParser
from djangorestframework_camel_case.render import CamelCaseJSONRenderer, CamelCaseBrowsableAPIRenderer

from django_microservice_common.api.exception_handlers import errors_exception_handler


class ApiViewV3Mixin:
    renderer_classes = (
        CamelCaseJSONRenderer,
        CamelCaseBrowsableAPIRenderer,
    )
    parser_classes = (
        CamelCaseFormParser,
        CamelCaseMultiPartParser,
        CamelCaseJSONParser,
    )

    def get_exception_handler(self):
        return errors_exception_handler



from typing import Sequence, Type, TYPE_CHECKING

from importlib import import_module

from django.conf import settings

from django.contrib import auth

from rest_framework.permissions import IsAuthenticated, BasePermission
from rest_framework.authentication import BaseAuthentication

from rest_framework_simplejwt.authentication import JWTAuthentication 


def get_auth_header(headers):
    value = headers.get('Authorization')

    if not value:
        return None

    auth_type, auth_value = value.split()[:2]

    return auth_type, auth_value


if TYPE_CHECKING:
    # This is going to be resolved in the stub library
    # https://github.com/typeddjango/djangorestframework-stubs/
    from rest_framework.permissions import _PermissionClass

    PermissionClassesType = Sequence[_PermissionClass]
else:
    PermissionClassesType = Sequence[Type[BasePermission]]


class ApiAuthMixin:
    authentication_classes: Sequence[Type[BaseAuthentication]] = [
            JWTAuthentication,
    ]
    permission_classes: PermissionClassesType = (IsAuthenticated, )



import uuid

from django.conf import settings

from ..authorization.grants import PermissionGrants, GRANT_MAPPER
from ..authorization.roles import UserRoles


def get_permission_setting(acl_mapper: dict[settings.RoutesEnum, dict[UserRoles, PermissionGrants]]):
    for router in acl_mapper:
        acl_mapper[router][UserRoles.ADMIN] = PermissionGrants.IS_ALLOWED
    return acl_mapper


def check_permission(
        user_rule: UserRoles,
        user_id: uuid.uuid4,
        route: settings.RoutesEnum,
        permission_setting: dict[settings.RoutesEnum, dict[UserRoles, PermissionGrants]],
        **kwargs,
):
    permission_for_endpoint = permission_setting.get(route)
    assert permission_for_endpoint is not None
    _permission_grant = permission_for_endpoint.get(user_rule)
    permission_grant = _permission_grant or PermissionGrants.NOT_ALLOWED
    permission_func = GRANT_MAPPER.get(permission_grant)
    assert permission_func is not None
    permission_func(user_id, **kwargs)


def get_permission_callable():
    return check_permission



from django.db import models
from django.utils import timezone


class BaseModel(models.Model):
    created_at = models.DateTimeField(db_index=True, default=timezone.now)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class ActivatableModel(models.Model):
    is_active = models.BooleanField(null=False, default=True)

    class Meta:
        abstract = True



from django.core.exceptions import ValidationError

from .roles import UserRoles


def validate_user_roles(value):
    if value not in UserRoles.values:
        raise ValidationError("Invalid roles!!")



from django.apps import AppConfig


class CommonConfig(AppConfig):
    default_auto_field = 'django.db.django_microservice_common.BigAutoField'
    name = "django_microservice_common"
    label = "django_microservice_common"
    verbose_name = "Common"



from rest_framework import status
from rest_framework.exceptions import APIException


class CustomApiException(APIException):
    ...


class NotAuthorizedError(CustomApiException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = "You are unauthorized"



from enum import auto

from ..authorization.enums import StrEnum
from ..authorization.grants_func import is_allowed, is_not_allowed


class PermissionGrants(StrEnum):
    IS_ALLOWED = auto()
    NOT_ALLOWED = auto()


GRANT_MAPPER = {
    PermissionGrants.IS_ALLOWED: is_allowed,
    PermissionGrants.NOT_ALLOWED: is_not_allowed,
}
